﻿XIncludeFile("PB64__DECLARE.pbi")

Macro CIA_DEBUG(text_)
	CompilerIf #Debug_CIA
		Debug "        CIA: " + text_
	CompilerEndIf
EndMacro

Macro CIA_SetOutput(p_, v_)
	p_\output_bits = ~(v_)
EndMacro

Macro CIA_SetInput(p_, v_)
	p_\input_bits = ~(v_)
EndMacro

Macro CIA_GetOutput(p_)
	~(p_\output_bits)
EndMacro

Macro CIA_GetInput(p_)
	~(p_\input_bits)
EndMacro

#CIA_TEST = 1

CompilerIf #CIA_TEST = 0
	Macro IS_FALSE(v_, b_)
		ISSET_BIT(v_, b_)
	EndMacro
	
	Macro IS_TRUE(v_, b_)
		(ISSET_BIT(v_, b_) = 0)
	EndMacro
	
	Macro SET_FALSE(v_, b_)
		v_ | (1 << (b_))
	EndMacro
	
	Macro SET_TRUE(v_, b_)
		v_ & ~(1 << (b_))
	EndMacro
CompilerElse
	Macro IS_FALSE(v_, b_)
		(ISSET_BIT(v_, b_) = 0)
	EndMacro
	
	Macro IS_TRUE(v_, b_)
		ISSET_BIT(v_, b_)
	EndMacro
	
	Macro SET_TRUE(v_, b_)
		v_ | (1 << (b_))
	EndMacro
	
	Macro SET_FALSE(v_, b_)
		v_ & ~(1 << (b_))
	EndMacro
CompilerEndIf

Procedure CIA_Init(*cia.S_CIA)
	ClearStructure(*cia, S_CIA)
	InitializeStructure(*cia, S_CIA)
	
	*cia\timer_a\latch				= 0
	*cia\timer_b\latch				= 0
	*cia\timer_a\counter			= 0
	*cia\timer_b\counter			= 0
	*cia\timer_a\enabled			= #False
	*cia\timer_b\enabled			= #False
	*cia\timer_a\irq_enabled		= #False
	*cia\timer_b\irq_enabled		= #False
	*cia\timer_a\irq_triggered		= #False
	*cia\timer_b\irq_triggered		= #False
	*cia\timer_a\input_mode			= #kModeSystemFrq
	*cia\timer_b\input_mode			= #kModeSystemFrq
	*cia\timer_a\run_mode			= #kModeRestart
	*cia\timer_b\run_mode 			= #kModeRestart
	*cia\pra						= $FF
	*cia\prb						= $FF
EndProcedure

Procedure PrintN2(t.s)
	AddGadgetItem(1, -1, t)
	SetGadgetState(1, CountGadgetItems(1) - 1)
EndProcedure
Procedure CIA_WriteRegister(*cia.S_CIA, r.a, v.a)
	r & $0F
	
	*cia\RAM[r] = v
	
	Select r
		Case $0
			If *cia = *CIA1
				*cia\pra = v
				; Data port a (PRA), keyboard matrix cols And joystick #2
			ElseIf *cia = *CIA2
				; Data port a (PRA)
				VIC\Mem_Base = ~(v & 3) << 14
				*cia\pra = v
			EndIf
		Case $1
			*cia\prb = v; | ~*cia\ddrb
			If *cia = *CIA1
				; Data port b (PRB), keyboard matrix rows And joystick #1
			ElseIf *cia = *CIA2
				; Data port b (PRB)
			EndIf
		Case $2
			; Data direction port a (DDRA)
			*cia\ddra = v
		Case $3
			; Data direction port b (DDRB)
			*cia\ddrb = v
		Case $4
			; timer a low byte
			*cia\timer_a\latch & $FF00
			*cia\timer_a\latch | v
		Case $5
			; timer a high byte
			*cia\timer_a\latch & $00FF
			*cia\timer_a\latch | (v << 8)
		Case $6
			; timer b low byte
			*cia\timer_b\latch & $FF00
			*cia\timer_b\latch | v
		Case $7
			; timer b high byte
			*cia\timer_b\latch & $00FF
			*cia\timer_b\latch | (v << 8)
		Case $8
			; RTC 1/10s
		Case $9
			; RTC seconds
		Case $A
			; RTC minutes
		Case $B
			; RTC hours
		Case $C
			; shift serial
		Case $D
			; interrupt control And status
			; If bit 7 is set, enable selected mask of 
			; interrupts, Else disable them
			If (ISSET_BIT(v, 0)) : *cia\timer_a\irq_enabled = ISSET_BIT(v, 7) : EndIf
			If (ISSET_BIT(v, 1)) : *cia\timer_b\irq_enabled = ISSET_BIT(v, 7) : EndIf
		Case $E
			; control timer a
			*cia\timer_a\enabled	= Bool(v & 1)
			*cia\timer_a\input_mode	= (v & $20) >> 5
			; load latch requested
			If v & (1 << 4)
				*cia\timer_a\counter = *cia\timer_a\latch
			EndIf
		Case $F
			; control timer b
			*cia\timer_b\enabled = Bool(v & $1)
			*cia\timer_b\input_mode = (v & $60) >> 5
			; load latch requested
			If (v & (1 << 4))
				*cia\timer_b\counter = *cia\timer_b\latch
			EndIf
	EndSelect
EndProcedure

Procedure.a CIA_ReadRegister(*cia.S_CIA, r.a)
	r & $0F
	
	Protected retval.a = *cia\RAM[r]
	
	Protected col.a, v.a
	Select r
		Case $00
			If *cia = *CIA1
				; Data port a (PRA), keyboard matrix cols And joystick #2
				If IO\InputMode = #C64_Joystick_Port2
					retval = ~IO\Joystick[#C64_Joystick_Port2]
				Else
					retval = $FF
				EndIf
			Else
				; Data port a (PRA)
				retval = *cia\pra
			EndIf
		Case $01
			; Data port b (PRB), keyboard matrix rows And joystick #1
			If *cia = *CIA1
				If IO\InputMode = #C64_KeyBrd
					If *cia\pra = $FF
						retval = $FF
					ElseIf *cia\pra
						col	= 0
						v	= ~(*cia\pra)
						While (v > 1)
							v >> 1
							col + 1
						Wend
						retval = IO\Keyboard_Matrix[col]
					EndIf
				ElseIf IO\InputMode = #C64_Joystick_Port1
					retval = ~IO\Joystick[#C64_Joystick_Port1]
				Else
					retval = $FF
				EndIf
			Else
				; Data port b (PRB)
				retval = *cia\prb
			EndIf
		Case $02  
			; Data direction port a (DDRA)
			retval = *cia\ddra
		Case $03    
			; Data direction port b (DDRB)
			retval = *cia\ddrb
		Case $04
			; timer a low byte
			retval = *cia\timer_a\counter & $00FF
		Case $05
			; timer a high byte
			retval = (*cia\timer_a\counter & $FF00) >> 8
		Case $06
			; timer b low byte
			retval = *cia\timer_b\counter & $00FF
		Case $07 
			; timer b high byte
			retval = (*cia\timer_b\counter & $ff00) >> 8
		Case $08
			; RTC 1/10s
		Case $09
			; RTC seconds
		Case $0A
			; RTC minutes
		Case $0B
			; RTC hours
		Case $0C
			; shift serial
		Case $0D
			; interrupt control And status
			If *cia\timer_a\irq_triggered Or *cia\timer_b\irq_triggered
				retval | (1 << 7); IRQ occured
				If *cia\timer_a\irq_triggered : retval | (1 << 0) : EndIf
				If *cia\timer_b\irq_triggered : retval | (1 << 1) : EndIf
			EndIf
			
			If *cia = *CIA1
				CPU\IRQ_State & ~#CIA_IRQ
			Else
				CPU\NMI_State & ~#CIA_NMI
			EndIf

			retval = 0;| %000010000
		Case $0E
			; control timer a
		Case $0F
			; control timer b
	EndSelect
	
	ProcedureReturn retval
EndProcedure

Procedure CIA_UpddateTimer(*cia, *timer.S_CIA_TIMER, count)
	If *timer\enabled
		Select *timer\input_mode
			Case #kModeSystemFrq
				*timer\counter - count
				If *timer\counter <= 0
					If *timer\irq_enabled
						*timer\irq_triggered = #True
						If *cia = *CIA1
							CPU\IRQ_State | #CIA_IRQ
						Else
							CPU\NMI_State | #CIA_NMI
						EndIf
					EndIf
					
					Select *timer\run_mode
						Case #kModeRestart:	*timer\counter = *timer\latch
						Case #kModeOneTime:	*timer\enabled = #False
					EndSelect
				EndIf
			Case #kModeCNT
			Case #kModeTimerA
			Case #kModeTimerACNT
		EndSelect
	EndIf
EndProcedure

Procedure CIA_Emulate(*cia.S_CIA, count)
	CIA_UpddateTimer(*cia, *cia\timer_a, count)
	CIA_UpddateTimer(*cia, *cia\timer_b, count)
	
; 	If *cia = *CIA2
; 		CIA_Test(*cia, *cia\wire)
; 	EndIf
	
	ProcedureReturn #True
EndProcedure
; IDE Options = PureBasic 5.72 (Windows - x64)
; CursorPosition = 183
; FirstLine = 178
; Folding = ----
; EnableXP