﻿XIncludeFile("PB64__DECLARE.pbi")

Macro CPU_Set_ZF(v_)
 	CPU\FLAG_Z = Bool(((v_) & $FF) = 0)
EndMacro

Macro CPU_Set_NF(v_)
	CPU\FLAG_N = Bool((v_) & $80)
EndMacro

Macro CPU_Set_NZ(v_)
	CPU\FLAG_N = Bool((v_) & $80)
	CPU\FLAG_Z = Bool(((v_) & $FF) = 0)	
EndMacro

Procedure.a CPU_NextOpcode()
	Protected opcode.a = MEMORY_Peek(CPU\PC)
	
	CompilerIf #Debug_CPU
		DBG\Param2 + RSet(Hex(opcode), 2, "0") + " "
	CompilerEndIf
	
	CPU\PC + 1
	ProcedureReturn opcode
EndProcedure

Procedure.u CPU_NextOpcodeW()
 	Protected opcode.u = MEMORY_ReadW(CPU\PC)
	
 	CPU\PC + 2
	ProcedureReturn opcode
EndProcedure

Procedure.a Addr_Immediate(checkBoundary.a)
	Protected ofs.a = CPU_NextOpcode()
	
	CompilerIf #Debug_CPU
		DBG\Param1 = "#" + RSet(Hex(ofs), 2, "0")
	CompilerEndIf

	ProcedureReturn ofs
EndProcedure

Procedure.a Addr_Zeropage(checkBoundary.a)
	Protected ofs.a = CPU_NextOpcode()
	
	CompilerIf #Debug_CPU
		DBG\Param1 = RSet(Hex(ofs), 2, "0")
	CompilerEndIf
	
	ProcedureReturn ofs
EndProcedure

Procedure.a Addr_ZeropageX(checkBoundary.a)
	Protected ofs.a = CPU_NextOpcode()

	CompilerIf #Debug_CPU
		DBG\Param1 = RSet(Hex(ofs), 2, "0") + ",x"
	CompilerEndIf
		
	ProcedureReturn ofs + CPU\X
EndProcedure

Procedure.a Addr_ZeropageY(checkBoundary.a)
	Protected ofs.a = CPU_NextOpcode()
	
	CompilerIf #Debug_CPU
		DBG\Param1 = RSet(Hex(ofs), 2, "0") + ",y"
	CompilerEndIf

	ProcedureReturn ofs + CPU\Y
EndProcedure

Procedure.u Addr_Absolute(checkBoundary.a)
	Protected ofs.u = CPU_NextOpcodeW()
	
	CompilerIf #Debug_CPU
		DBG\Param1 = Hex(ofs)
	CompilerEndIf
	
	ProcedureReturn ofs
EndProcedure

Procedure.u Addr_AbsoluteX(checkBoundary.a)
	Protected ofs.u = CPU_NextOpcodeW()

	CompilerIf #Debug_CPU
		DBG\Param1 = RSet(Hex(ofs), 4, "0") + ",x"
	CompilerEndIf
	
	If checkBoundary
		If ((ofs & $FF) + CPU\X) > $FF
			CPU\Cycles + 1			
			Protected lowByte.a = (ofs + CPU\X)
			MEMORY_Peek((ofs & $FF00) | lowByte)
;  			ProcedureReturn ofs
		EndIf
	EndIf
	
	ProcedureReturn ofs + CPU\X
EndProcedure

Procedure.u Addr_AbsoluteY(checkBoundary.a)
	Protected ofs.u = CPU_NextOpcodeW()
	
	CompilerIf #Debug_CPU
		DBG\Param1 = RSet(Hex(ofs), 4, "0") + ",y"
	CompilerEndIf	
	
	If checkBoundary
		If ((ofs & $FF) + CPU\Y) > $FF
			CPU\Cycles + 1
			Protected lowByte.a = (ofs + CPU\Y)
			MEMORY_Peek((ofs & $FF00) | lowByte)
		EndIf
	EndIf

	
	ProcedureReturn ofs + CPU\Y
EndProcedure

Procedure.u Addr_Indirect(checkBoundary.a)
	Protected op.u = CPU_NextOpcodeW()
	Protected ofs.u = MEMORY_ReadW(op)

	CompilerIf #Debug_CPU
		DBG\Param1 = "(" + RSet(Hex(op), 4, "0") + ")"
	CompilerEndIf	
	
	ProcedureReturn ofs
EndProcedure

Procedure.u Addr_IndirectX(checkBoundary.a)
	Protected op.a = CPU_NextOpcode()
 	Protected ofs.u = MEMORY_ReadW((op + CPU\X) & $FF)

	CompilerIf #Debug_CPU
		DBG\Param1 = "(" + RSet(Hex(op), 2, "0") + ",x)"
	CompilerEndIf	
	
	ProcedureReturn ofs
EndProcedure

Procedure.u Addr_IndirectY(checkBoundary.a)
	Protected op.a = CPU_NextOpcode()
  	Protected ofs.u = MEMORY_ReadW(op)
  	
	CompilerIf #Debug_CPU
		DBG\Param1 = "(" + RSet(Hex(op), 2, "0") + "),y"
	CompilerEndIf

	
	If checkBoundary
		If ((ofs & $FF) + CPU\Y) > $FF
			CPU\Cycles + 1
			Protected lowByte.a = (ofs + CPU\Y)
			MEMORY_Peek((ofs & $FF00) | lowByte)	
		EndIf
	EndIf
	
	ProcedureReturn ofs + CPU\Y
EndProcedure

Procedure.u Addr_Relative(checkBoundary.a)	
	Protected op.b = CPU_NextOpcode()
  	Protected ofs.u = CPU\PC + op
  	  	
	CompilerIf #Debug_CPU
		DBG\Param1 = RSet(Hex(ofs), 4, "0")
	CompilerEndIf
	
	ProcedureReturn ofs
EndProcedure

Procedure CPU_Push(v.a)
	Protected ofs.u = #AddressStack + CPU\SP
	MEMORY_Poke(ofs, v)
	CPU\SP - 1
EndProcedure

Procedure.a CPU_Pop()
	CPU\SP + 1
	Protected ofs.u = #AddressStack + CPU\SP
	ProcedureReturn MEMORY_Peek(ofs)
EndProcedure

Procedure.a CPU_GetFlags()
	Protected flags.a = (CPU\FLAG_C << 0) |
	                    (CPU\FLAG_Z << 1) | 
	                    (CPU\FLAG_I << 2) | 
	                    (CPU\FLAG_D << 3) | 
	                    (1 << 4) | 
	                    (1 << 5) | 
	                    (CPU\FLAG_O << 6) | 
	                    (CPU\FLAG_N << 7)
	ProcedureReturn flags
EndProcedure

Procedure CPU_SetFlags(v.a)
	CPU\FLAG_C = ISSET_BIT(v, 0)
	CPU\FLAG_Z = ISSET_BIT(v, 1)
	CPU\FLAG_I = ISSET_BIT(v, 2)
	CPU\FLAG_D = ISSET_BIT(v, 3)
	CPU\FLAG_O = ISSET_BIT(v, 6)
	CPU\FLAG_N = ISSET_BIT(v, 7)
EndProcedure

Procedure CPU_Reset()
	CPU\FLAG_C	= 0
	CPU\FLAG_Z	= 0
	CPU\FLAG_I	= 0
	CPU\FLAG_D	= 0
	CPU\FLAG_B	= 1
	CPU\FLAG_O	= 0
	CPU\FLAG_N	= 0
	CPU\A		= 0
	CPU\X		= 0
	CPU\Y		= 0
	CPU\SP		= 0
	CPU\PC		= MEMORY_ReadW(#AddressColdReset)
	
	CPU\Cycles = 6
EndProcedure


Procedure CPU_IRQ()
	If (CPU\FLAG_I = 0)
		CPU_Push((CPU\PC >> 8) & $FF)
		CPU_Push((CPU\PC       & $FF))
		CPU_Push(CPU_GetFlags() & %11101111)
		CPU\PC = MEMORY_ReadW(#AddressIRQ)
		CPU\FLAG_I = 1
		CPU\Cycles + 7
	EndIf
EndProcedure

Procedure CPU_NMI()
	CPU_Push((CPU\PC >> 8) & $FF)
	CPU_Push( CPU\PC       & $FF)
	CPU_Push((CPU_GetFlags() & %11101111))
	CPU\PC = MEMORY_ReadW(#AddressNMI)
	CPU\Cycles + 7
EndProcedure

Procedure BRK()
	CPU_Push(((CPU\PC + 1) >> 8) & $FF)
	CPU_Push(( CPU\PC + 1)       & $FF)
	
	CPU\FLAG_B = 1
	CPU_Push(CPU_GetFlags())
	CPU\FLAG_I = 1
	
	CPU\PC = MEMORY_ReadW(#AddressIRQ)
EndProcedure

Procedure.a ASL(v.a)
	Protected t.u = (v << 1)
	
	CPU\FLAG_C = Bool(v & $80)
	CPU_Set_NZ(t)
	
	ProcedureReturn t
EndProcedure

Procedure ASL_A()
	CPU\A = ASL(CPU\A)
EndProcedure

Procedure ASL_M(ofs.u)
	Protected v.a = MEMORY_Peek(ofs)
	
	MEMORY_Poke(ofs, v)
	MEMORY_Poke(ofs, ASL(v))
EndProcedure

Procedure AND_(v.a)
	CPU\A & v
	CPU_Set_NZ(CPU\A)
EndProcedure

Procedure BCC(ofs.u)
	If CPU\FLAG_C = 0
		If (CPU\PC & $F000) <> (ofs & $F000)
			CPU\Cycles + 1
		EndIf

		CPU\PC = ofs
		CPU\Cycles + 1
	EndIf
EndProcedure  

Procedure BCS(ofs.u)
	If CPU\FLAG_C
		If (CPU\PC & $F000) <> (ofs & $F000)
			CPU\Cycles + 1
		EndIf
		
		CPU\PC = ofs
		CPU\Cycles + 1
	EndIf
EndProcedure  

Procedure BEQ(ofs.u)
	If CPU\FLAG_Z
		If (CPU\PC & $F000) <> (ofs & $F000)
			CPU\Cycles + 1
		EndIf
		
		CPU\PC = ofs
		CPU\Cycles + 1
	EndIf
EndProcedure

Procedure BMI(ofs.u)
	If CPU\FLAG_N
		If (CPU\PC & $F000) <> (ofs & $F000)
			CPU\Cycles + 1
		EndIf
		
		CPU\PC = ofs
		CPU\Cycles + 1
	EndIf
EndProcedure

Procedure BNE(ofs.u)
	If CPU\FLAG_Z = 0
		If (CPU\PC & $F000) <> (ofs & $F000)
			CPU\Cycles + 1
		EndIf
		
		CPU\PC = ofs
		CPU\Cycles + 1
	EndIf
EndProcedure

Procedure BPL(ofs.u)
	If CPU\FLAG_N = 0
		If (CPU\PC & $F000) <> (ofs & $F000)
			CPU\Cycles + 1
		EndIf
		
		CPU\PC = ofs
		CPU\Cycles + 1
	EndIf
EndProcedure

Procedure BVC(ofs.u)
	If CPU\FLAG_O = 0
		If (CPU\PC & $F000) <> (ofs & $F000)
			CPU\Cycles + 1
		EndIf
		
		CPU\PC = ofs
		CPU\Cycles + 1
	EndIf
EndProcedure

Procedure BVS(ofs.u)
	If CPU\FLAG_O
		If (CPU\PC & $F000) <> (ofs & $F000)
			CPU\Cycles + 1
		EndIf
		
		CPU\PC = ofs
		CPU\Cycles + 1
	EndIf
EndProcedure

Procedure BIT(ofs.u)
	Protected t.a = MEMORY_Peek(ofs)
	
	CPU\FLAG_O = Bool(t & $40)
	CPU_Set_NF(t)
	CPU_Set_ZF(t & CPU\A)
EndProcedure

Procedure CMP(v.a)
	Protected t.u = CPU\A - v
	
	CPU\FLAG_C = Bool(t < $100)
	CPU_Set_NZ(t)
EndProcedure

Procedure CPX(v.a)
	Protected t.u = CPU\X - v
	
	CPU\FLAG_C = Bool(t < $100)
	t & $FF
	CPU_Set_NZ(t)
EndProcedure

Procedure CPY(v.a)
	Protected t.u = CPU\Y - v
	
	CPU\FLAG_C = Bool(t < $100)	
	t & $FF	
	CPU_Set_NZ(t)
EndProcedure

Procedure DEC(ofs.u)
	Protected v.a = MEMORY_Peek(ofs)
	
	MEMORY_Poke(ofs, v)
	v - 1
	MEMORY_Poke(ofs, v)
	CPU_Set_NZ(v)
EndProcedure

Procedure DEX()
	CPU\X - 1
	CPU_Set_NZ(CPU\X)
EndProcedure

Procedure DEY()
	CPU\Y - 1
	CPU_Set_NZ(CPU\Y)
EndProcedure

Procedure EOR(v.a)
	CPU\A ! v
	CPU_Set_NZ(CPU\A)
EndProcedure

Procedure INC(ofs.u)
	Protected v.a = MEMORY_Peek(ofs)
	
	MEMORY_Poke(ofs, v)
	v + 1
	MEMORY_Poke(ofs, v)
	CPU_Set_NZ(v)
EndProcedure

Procedure INX()
	CPU\X + 1
	CPU_Set_NZ(CPU\X)
EndProcedure

Procedure INY()
	CPU\Y + 1
	CPU_Set_NZ(CPU\Y)
EndProcedure

Procedure JMP(ofs.u)
	CPU\PC = ofs
EndProcedure

Procedure JSR(ofs.u)
	CPU_Push(((CPU\PC - 1) >> 8) & $FF)
	CPU_Push( (CPU\PC - 1)       & $FF)
	CPU\PC = ofs
EndProcedure

Procedure LDA(v.a)
	CPU\A = v
	CPU_Set_NZ(CPU\A)
EndProcedure

Procedure LDX(v.a)
	CPU\X = v
	CPU_Set_NZ(CPU\X)
EndProcedure

Procedure LDY(v.a)
	CPU\Y = v
	CPU_Set_NZ(CPU\Y)
EndProcedure

Procedure.a LSR(v.a)
	Protected t.a = v >> 1
	
	CPU\FLAG_C = Bool(v & 1)
	CPU_Set_NZ(t)
	
	ProcedureReturn t
EndProcedure

Procedure LSR_A()
	CPU\A = LSR(CPU\A)
EndProcedure

Procedure LSR_M(ofs.u)
	Protected v.a = MEMORY_Peek(ofs)
	
	MEMORY_Poke(ofs, v)
	MEMORY_Poke(ofs, LSR(v))
EndProcedure

Procedure NOP()
EndProcedure

Procedure ORA(v.a)
	CPU\A = (CPU\A | v)
 	CPU_Set_NZ(CPU\A)
EndProcedure


Procedure PHA()
	CPU_Push(CPU\A)
EndProcedure

Procedure PHP()
	CPU_Push(CPU_GetFlags())
EndProcedure

Procedure PLA()
	CPU\A = CPU_Pop()
	CPU_Set_NZ(CPU\A)
EndProcedure

Procedure PLP()
	CPU_SetFlags(CPU_Pop()) 	
EndProcedure

Procedure.a ROL(v.a)
	Protected t.u = (v << 1) + CPU\FLAG_C
	
	CPU\FLAG_C = Bool(t & $100)
	CPU_Set_NZ(t)
	
	ProcedureReturn t
EndProcedure

Procedure ROL_A()
	CPU\A = ROL(CPU\A)
EndProcedure

Procedure ROL_M(ofs.u)
	Protected v.a = MEMORY_Peek(ofs)
	
	MEMORY_Poke(ofs, v)
	MEMORY_Poke(ofs, ROL(v))
EndProcedure

; Rotate one bit Right (memory Or accumulator)
Procedure.a ROR(v.a)
	Protected t.u = (v >> 1) + (CPU\FLAG_C << 7)
	
	CPU\FLAG_C = (v & $1)
	CPU_Set_NZ(t)
	
	ProcedureReturn t
EndProcedure

Procedure ROR_A()
	CPU\A = ROR(CPU\A)
EndProcedure

Procedure ROR_M(ofs.u)
	Protected v.a = MEMORY_Peek(ofs)
	
	MEMORY_Poke(ofs, v)
	MEMORY_Poke(ofs, ROR(v))
EndProcedure

Procedure RTI()
	CPU_SetFlags(CPU_Pop())
	CPU\PC = CPU_Pop() | (CPU_Pop() << 8)
EndProcedure

Procedure RTS()
	Protected ofs.u = (CPU_Pop() | (CPU_Pop() << 8)) + 1
	CPU\PC = ofs
EndProcedure

Procedure ADC(v.a)
	Protected t.u
	
	If CPU\FLAG_D
 		t = (CPU\A & $F) + (v & $F) + CPU\FLAG_C
		If (t > $09)
			t + $06
		EndIf
		t + ((CPU\A & $F0) + (v & $F0))
		If ((t & $1F0) > $90)
			t + $60
		EndIf
	Else
		t = CPU\A + v + CPU\FLAG_C
	EndIf
	
	CPU\FLAG_C = Bool(t > $FF)
	t & $FF
	
	If (((CPU\A ! v) & $80) = 0) And ((CPU\A ! t) & $80)
		CPU\FLAG_O = 1
	Else
		CPU\FLAG_O = 0
	EndIf
	
	CPU_Set_NZ(t)
	
	CPU\A = t
EndProcedure

Procedure SBC(v.a)
	Protected t.u
	
	If CPU\FLAG_D
		t = (CPU\A & $F) - (v & $F) + (CPU\FLAG_C - 1)
		
		If (t & $10)
			t = ((t - $6) & $F) | ((CPU\A & $F0) - (v & $F0) - $10)
		Else
			t = (t & $F) | ((CPU\A & $F0) - (v & $F0))
		EndIf
		If (t & $100)
			t - $60
		EndIf
	Else
		t = CPU\A - v + (CPU\FLAG_C - 1)
	EndIf
	
	CPU\FLAG_C = Bool(t < $100)
	t & $FF
	
	If ((CPU\A ! t) & $80) And ((CPU\A ! v) & $80)
		CPU\FLAG_O = 1
	Else
		CPU\FLAG_O = 0
	EndIf
	
	CPU_Set_NZ(t)
	
	CPU\A = t
EndProcedure

Procedure SEI()
	CPU\FLAG_I = 1
EndProcedure

Procedure CLI()
	CPU\FLAG_I = 0
EndProcedure

Procedure SEC()
	CPU\FLAG_C = 1
EndProcedure

Procedure CLC()
	CPU\FLAG_C = 0
EndProcedure

Procedure SED()
	CPU\FLAG_D = 1
EndProcedure

Procedure CLD()
	CPU\FLAG_D = 0
EndProcedure

Procedure CLV()
	CPU\FLAG_O = 0
EndProcedure

Procedure STA(ofs.u)
	MEMORY_Poke(ofs, CPU\A)
EndProcedure

Procedure STX(ofs.u)
	MEMORY_Poke(ofs, CPU\X)
EndProcedure

Procedure STY(ofs.u)
	MEMORY_Poke(ofs, CPU\Y)
EndProcedure

Procedure TAX()
	CPU\X = CPU\A
	CPU_Set_NZ(CPU\X)
EndProcedure

Procedure TAY()
	CPU\Y = CPU\A
	CPU_Set_NZ(CPU\Y)
EndProcedure

Procedure TSX()
	CPU\X = CPU\SP
	CPU_Set_NZ(CPU\X)
EndProcedure

Procedure TXA()
	CPU\A = CPU\X
	CPU_Set_NZ(CPU\A)
EndProcedure

Procedure TXS()
	CPU\SP = CPU\X
EndProcedure
	
Procedure TYA()
	CPU\A = CPU\Y
	CPU_Set_NZ(CPU\A)
EndProcedure

;**************************************************
;-   ILLEGAL OPCODES
;**************************************************

Procedure KIL()
	C64\State | #C64_State_Error
	C64\ErrorMessage	= "CPU: Jam occured."
EndProcedure

Procedure SLO(ofs.u)
	ASL_M(ofs)
	ORA(MEMORY_Peek(ofs))
EndProcedure

Procedure SAX(ofs.u)
	Protected v.a = CPU\A & CPU\X
	MEMORY_Poke(ofs, v)
EndProcedure

Procedure LAS(ofs.u)
	CPU\SP & MEMORY_Peek(ofs)
	CPU\X = CPU\SP
	CPU\A = CPU\SP
	CPU_Set_NZ(CPU\SP)
EndProcedure

Procedure RRA(ofs.u)
 	ROR_M(ofs)
 	ADC(MEMORY_Peek(ofs))
EndProcedure

Procedure DCP(ofs.u)
	DEC(ofs)
	CMP(MEMORY_Peek(ofs))
EndProcedure

Procedure AXA(ofs.u)
	Protected v.a = CPU\A & CPU\X
	v & ((ofs >> 8) + 1)
	MEMORY_Poke(ofs, v)
EndProcedure

Procedure ISC(ofs.u)
	INC(ofs)
	SBC(MEMORY_Peek(ofs))
EndProcedure

Procedure RLA(ofs.u)
 	ROL_M(ofs)
	AND_(MEMORY_Peek(ofs))
EndProcedure

Procedure ANC(v.a)
	AND_(v)
	CPU\FLAG_C = CPU\FLAG_N
EndProcedure

Procedure SRE(v.a)
	LSR(v)
	EOR(v)
EndProcedure

Procedure ALR(ofs.u)
	AND_(MEMORY_Peek(ofs))
	LSR(CPU\A)
EndProcedure

Procedure ARR(ofs.u)
	AND_(MEMORY_Peek(ofs))
	ROR(CPU\A)
EndProcedure

Procedure XAA(v.a)
	TAX()
	AND_(v)
EndProcedure

Procedure TAS(ofs.u)
	Protected v.a = CPU\A & CPU\X
	CPU\SP = v
	v & ((ofs >> 8) + 1)
	MEMORY_Poke(ofs, v)
EndProcedure

Procedure SHY(ofs.u)
	Protected v.a = CPU\Y & MEMORY_Peek((ofs >> 8) + 1)
	MEMORY_Poke(ofs, v)
EndProcedure

Procedure SHX(ofs.u)
	Protected v.a = CPU\X & MEMORY_Peek((ofs >> 8) + 1)
	MEMORY_Poke(ofs, v)
EndProcedure

Procedure LAX(ofs.u)
	LDA(MEMORY_Peek(ofs))
	LDX(MEMORY_Peek(ofs))
EndProcedure

;-

Procedure CPU_Emulate()
	CompilerIf #Debug_CPU
		ClearStructure(DBG, S_DEBUG)
	CompilerEndIf	
	
	Protected lastCycle = CPU\Cycles
	
	Static NMIStateOld	
	If CPU\NMI_State And (NMIStateOld = 0)
		CPU_NMI()
 	ElseIf CPU\IRQ_State
 		CPU_IRQ()
 	EndIf
 	NMIStateOld = CPU\NMI_State	
	
	Protected pc.u	= CPU\PC
	Protected op.a	= CPU_NextOpcode()
	
	If Opcode(op)\addrMode = #Addr_Implied
		CallFunctionFast(Opcode(op)\proc)
	Else
		Protected ofs.u = CallFunctionFast(Opcode(op)\addrProc, Opcode(op)\boundaryCheck)

		If Opcode(op)\readByte
			CallFunctionFast(Opcode(op)\proc, MEMORY_Peek(ofs))
		Else
			CallFunctionFast(Opcode(op)\proc, ofs)
		EndIf
	EndIf
	
	CPU\Cycles + Opcode(op)\nCycles
	
	CompilerIf #Debug_CPU
		Debug RSet(Hex(pc), 4, "0") + ":    " + Opcode(op)\name + " " + DBG\Param1
	CompilerEndIf
		
	ProcedureReturn CPU\Cycles - lastCycle
EndProcedure

Procedure CPU_Debug(startOfs.u, lastOfs.u, outFile.s, memSetup.a = $FF)
	Protected pc, illegal.s
	Protected file = CreateFile(#PB_Any, outFile)
	
	If IsFile(file)
		If memSetup <> $FF
			MEMORY_Poke(#AddressMemoryConfig, memSetup)
		EndIf
		
		pc = startOfs
		
		WriteStringN(file, RSet("Addr", 4) + ":  " + LSet("B1 B2 B3", 10) + LSet("CY",4) + "OPC" + " " + "PARAM")
		WriteStringN(file, RSet("", 32, "-"))
		
		Repeat
			ClearStructure(DBG, S_DEBUG)
			
			CPU\PC = pc
			Protected op.a = CPU_NextOpcode()
			Protected size = 1
			
			If Opcode(op)\addrMode <> #Addr_Implied
				CallFunctionFast(Opcode(op)\addrProc)
				size + Opcode(op)\nBytes
			EndIf
			
			If Opcode(op)\isIllegal
				illegal = "*"
			Else
				illegal = " "
			EndIf
			
			WriteStringN(file, RSet(Hex(pc), 4, "0") + ":  " + LSet(DBG\Param2, 10) + LSet(Str(Opcode(op)\nCycles),3) + illegal + Opcode(op)\name + " " + DBG\Param1)
			
			pc + size
		Until pc > lastOfs
		
		CloseFile(file)
	EndIf
EndProcedure

Procedure CPU_Dump(startOfs.u, lastOfs.u, outFile.s)
	Protected pc
	Protected file = CreateFile(#PB_Any, outFile)
	
	If IsFile(file)
		For pc = startOfs To lastOfs
 			WriteByte(file, MEMORY_Peek(pc))
			;WriteByte(file, MEMORY\RAM[pc])
		Next
		
		CloseFile(file)
	EndIf
EndProcedure

Procedure CPU_AddCommand(index, name.s, *proc, addrMode, nCycles, boundaryCheck, isIllegal)
	Opcode(index)\name			= name
	Opcode(index)\proc			= *proc
	Opcode(index)\addrMode		= addrMode
	Opcode(index)\nCycles		= nCycles	
	Opcode(index)\boundaryCheck	= boundaryCheck
	Opcode(index)\isIllegal		= isIllegal
	
	Select addrMode
		Case #Addr_Implied:			Opcode(index)\Addrproc		= #Null				: Opcode(index)\nBytes = 0
		Case #Addr_Immediate:		Opcode(index)\Addrproc		= @Addr_Immediate() : Opcode(index)\nBytes = 1
		Case #Addr_Relative:		Opcode(index)\Addrproc		= @Addr_Relative()	: Opcode(index)\nBytes = 1
			
		Case #Addr_Zeropage:		Opcode(index)\Addrproc		= @Addr_Zeropage()	: Opcode(index)\nBytes = 1
		Case #Addr_ZeropageX:		Opcode(index)\Addrproc		= @Addr_Zeropagex()	: Opcode(index)\nBytes = 1
		Case #Addr_ZeropageY:		Opcode(index)\Addrproc		= @Addr_ZeropageY()	: Opcode(index)\nBytes = 1
		Case #Addr_Absolute:		Opcode(index)\Addrproc		= @Addr_Absolute()	: Opcode(index)\nBytes = 2
		Case #Addr_AbsoluteX:		Opcode(index)\Addrproc		= @Addr_AbsoluteX()	: Opcode(index)\nBytes = 2
		Case #Addr_AbsoluteY:		Opcode(index)\Addrproc		= @Addr_AbsoluteY()	: Opcode(index)\nBytes = 2
		Case #Addr_Indirect:		Opcode(index)\Addrproc		= @Addr_Indirect()	: Opcode(index)\nBytes = 2
		Case #Addr_IndirectX:		Opcode(index)\Addrproc		= @Addr_IndirectX()	: Opcode(index)\nBytes = 1
		Case #Addr_IndirectY:		Opcode(index)\Addrproc		= @Addr_IndirectY()	: Opcode(index)\nBytes = 1
			
		Case #Addr_IZeropage:		Opcode(index)\Addrproc		= @Addr_Zeropage()	: Opcode(index)\nBytes = 1	:	Opcode(index)\readByte = #True
		Case #Addr_IZeropageX:		Opcode(index)\Addrproc		= @Addr_ZeropageX()	: Opcode(index)\nBytes = 1	:	Opcode(index)\readByte = #True
		Case #Addr_IZeropageY:		Opcode(index)\Addrproc		= @Addr_ZeropageY()	: Opcode(index)\nBytes = 1	:	Opcode(index)\readByte = #True
		Case #Addr_IAbsolute:		Opcode(index)\Addrproc		= @Addr_Absolute()	: Opcode(index)\nBytes = 2	:	Opcode(index)\readByte = #True
		Case #Addr_IAbsoluteX:		Opcode(index)\Addrproc		= @Addr_AbsoluteX()	: Opcode(index)\nBytes = 2	:	Opcode(index)\readByte = #True
		Case #Addr_IAbsoluteY:		Opcode(index)\Addrproc		= @Addr_AbsoluteY()	: Opcode(index)\nBytes = 2	:	Opcode(index)\readByte = #True
		Case #Addr_IIndirectX:		Opcode(index)\Addrproc		= @Addr_IndirectX()	: Opcode(index)\nBytes = 1	:	Opcode(index)\readByte = #True
		Case #Addr_IIndirectY:		Opcode(index)\Addrproc		= @Addr_IndirectY()	: Opcode(index)\nBytes = 1	:	Opcode(index)\readByte = #True
	EndSelect
EndProcedure

Procedure CPU_Init()
	ClearStructure(@CPU, S_CPU)
	InitializeStructure(@CPU, S_CPU)
		
	CPU_AddCommand($00, "BRK", @BRK(), #Addr_Implied,		7, 0, 0)
	CPU_AddCommand($01, "ORA", @ORA(), #Addr_IIndirectX,	6, 0, 0)
	CPU_AddCommand($02, "KIL", @KIL(), #Addr_Implied,		1, 0, 1); illegal
	CPU_AddCommand($03, "SLO", @SLO(), #Addr_IndirectX,		8, 0, 1); illegal
	CPU_AddCommand($04, "NOP", @NOP(), #Addr_Zeropage,		3, 0, 1); illegal
	CPU_AddCommand($05, "ORA", @ORA(), #Addr_IZeropage,		3, 0, 0)
	CPU_AddCommand($06, "ASL", @ASL_M(), #Addr_Zeropage,	5, 0, 0)
	CPU_AddCommand($07, "SLO", @SLO(), #Addr_Zeropage,		5, 0, 1); illegal
	CPU_AddCommand($08, "PHP", @PHP(), #Addr_Implied,		3, 0, 0)
	CPU_AddCommand($09, "ORA", @ORA(), #Addr_Immediate,		2, 0, 0)
	CPU_AddCommand($0A, "ASL", @ASL_A(), #Addr_Implied,		2, 0, 0)
	CPU_AddCommand($0B, "ANC", @ANC(), #Addr_Immediate, 	2, 0, 1); illegal
	CPU_AddCommand($0C, "NOP", @NOP(), #Addr_Absolute,		4, 0, 1); illegal
	CPU_AddCommand($0D, "ORA", @ORA(), #Addr_IAbsolute,		4, 0, 0)
	CPU_AddCommand($0E, "ASL", @ASL_M(), #Addr_Absolute,	6, 0, 0)
	CPU_AddCommand($0F, "SLO", @SLO(), #Addr_Absolute,		6, 0, 1); illegal
	CPU_AddCommand($10, "BPL", @BPL(), #Addr_Relative,		2, 0, 0)
	CPU_AddCommand($11, "ORA", @ORA(), #Addr_IIndirectY,	5, 1, 0)
	CPU_AddCommand($12, "KIL", @KIL(), #Addr_Implied,		1, 0, 1); illegal
	CPU_AddCommand($13, "SLO", @SLO(), #Addr_IndirectY,		8, 0, 1); illegal
	CPU_AddCommand($14, "NOP", @NOP(), #Addr_ZeropageX,		4, 0, 1); illegal
	CPU_AddCommand($15, "ORA", @ORA(), #Addr_IZeropageX,	4, 0, 0)
	CPU_AddCommand($16, "ASL", @ASL_M(), #Addr_ZeropageX,	6, 0, 0)
	CPU_AddCommand($17, "SLO", @SLO(), #Addr_ZeropageX,		6, 0, 1); illegal)
	CPU_AddCommand($18, "CLC", @CLC(), #Addr_Implied,		2, 0, 0)
	CPU_AddCommand($19, "ORA", @ORA(), #Addr_IAbsoluteY,	4, 1, 0)
	CPU_AddCommand($1A, "NOP", @NOP(), #Addr_Implied,		4, 0, 1); illegal
	CPU_AddCommand($1B, "SLO", @SLO(), #Addr_AbsoluteY,		7, 0, 1); illegal
	CPU_AddCommand($1C, "NOP", @NOP(), #Addr_AbsoluteX,		4, 0, 1); illegal
	CPU_AddCommand($1D, "ORA", @ORA(), #Addr_IAbsoluteX,	4, 1, 0)	
	CPU_AddCommand($1E, "ASL", @ASL_M(), #Addr_AbsoluteX,	7, 0, 0)
	CPU_AddCommand($1F, "SLO", @SLO(), #Addr_AbsoluteX,		7, 0, 1); illegal
	CPU_AddCommand($20, "JSR", @JSR(), #Addr_Absolute,		6, 0, 0)
	CPU_AddCommand($21, "AND", @AND_(), #Addr_IIndirectX,	6, 0, 0)
	CPU_AddCommand($22, "KIL", @KIL(), #Addr_Implied,		1, 0, 1); illegal
	CPU_AddCommand($23, "RLA", @RLA(), #Addr_IndirectX, 	8, 0, 1); illegal
	CPU_AddCommand($24, "BIT", @BIT(), #Addr_Zeropage,		3, 0, 0)
	CPU_AddCommand($25, "AND", @AND_(), #Addr_IZeropage,	3, 0, 0)
	CPU_AddCommand($26, "ROL", @ROL_M(), #Addr_Zeropage,	5, 0, 0)
	CPU_AddCommand($27, "RLA", @RLA(), #Addr_Zeropage,		5, 0, 1); illegal
	CPU_AddCommand($28, "PLP", @PLP(), #Addr_Implied,		4, 0, 0)
	CPU_AddCommand($29, "AND", @AND_(), #Addr_Immediate,	2, 0, 0)
	CPU_AddCommand($2A, "ROL", @ROL_A(), #Addr_Implied,		2, 0, 0)
	CPU_AddCommand($2B, "ANC", @ANC(), #Addr_Immediate, 	2, 0, 1); illegal
	CPU_AddCommand($2C, "BIT", @BIT(), #Addr_Absolute,		4, 0, 0)
	CPU_AddCommand($2D, "AND", @AND_(), #Addr_IAbsolute,	4, 0, 0)
	CPU_AddCommand($2E, "ROL", @ROL_M(), #Addr_Absolute,	6, 0, 0)
	CPU_AddCommand($2F, "RLA", @RLA(), #Addr_Absolute, 		6, 0, 1); illegal
	CPU_AddCommand($30, "BMI", @BMI(), #Addr_Relative,		2, 0, 0)
	CPU_AddCommand($31, "AND", @AND_(), #Addr_IIndirectY,	5, 1, 0)
	CPU_AddCommand($32, "KIL", @KIL(), #Addr_Implied,		1, 0, 1); illegal
	CPU_AddCommand($33, "RLA", @RLA(), #Addr_IndirectY, 	8, 0, 1); illegal
	CPU_AddCommand($34, "NOP", @NOP(), #Addr_ZeropageX,		4, 0, 1); illegal
	CPU_AddCommand($35, "AND", @AND_(), #Addr_IZeropageX,	4, 0, 0)
	CPU_AddCommand($36, "ROL", @ROL_M(), #Addr_ZeropageX,	6, 0, 0)	
	CPU_AddCommand($37, "RLA", @RLA(), #Addr_ZeropageX, 	6, 0, 1); illegal
	CPU_AddCommand($38, "SEC", @SEC(), #Addr_Implied,		2, 0, 0)
	CPU_AddCommand($39, "AND", @AND_(), #Addr_IAbsoluteY,	4, 1, 0)
	CPU_AddCommand($3A, "NOP", @NOP(), #Addr_Implied,		2, 0, 1); illegal
	CPU_AddCommand($3B, "RLA", @RLA(), #Addr_AbsoluteY, 	7, 0, 1); illegal
	CPU_AddCommand($3C, "NOP", @NOP(), #Addr_AbsoluteX,		4, 0, 1); illegal
	CPU_AddCommand($3D, "AND", @AND_(), #Addr_IAbsoluteX,	4, 1, 0)
	CPU_AddCommand($3E, "ROL", @ROL_M(), #Addr_AbsoluteX,	7, 0, 0)
	CPU_AddCommand($3F, "RLA", @RLA(), #Addr_AbsoluteX, 	7, 0, 1); illegal
	CPU_AddCommand($40, "RTI", @RTI(), #Addr_Implied,		6, 0, 0)
	CPU_AddCommand($41, "EOR", @EOR(), #Addr_IIndirectX,	6, 0, 0)
	CPU_AddCommand($42, "KIL", @KIL(), #Addr_Implied,		1, 0, 1); illegal
	CPU_AddCommand($43, "SRE", @SRE(), #Addr_IndirectX,		8, 0, 1); illegal
	CPU_AddCommand($44, "NOP", @NOP(), #Addr_Zeropage,		4, 0, 1); illegal
	CPU_AddCommand($45, "EOR", @EOR(), #Addr_IZeropage,		3, 0, 0)
	CPU_AddCommand($46, "LSR", @LSR_M(), #Addr_Zeropage,	5, 0, 0)
	CPU_AddCommand($47, "SRE", @SRE(), #Addr_Zeropage,		5, 0, 1); illegal
	CPU_AddCommand($48, "PHA", @PHA(), #Addr_Implied,		3, 0, 0)
	CPU_AddCommand($49, "EOR", @EOR(), #Addr_Immediate,		2, 0, 0)
	CPU_AddCommand($4A, "LSR", @LSR_A(), #Addr_Implied,		2, 0, 0)
	CPU_AddCommand($4B, "ALR", @ALR(), #Addr_Immediate,		2, 0, 1); illegal
	CPU_AddCommand($4C, "JMP", @JMP(), #Addr_Absolute,		3, 0, 0)
	CPU_AddCommand($4D, "EOR", @EOR(), #Addr_IAbsolute,		4, 0, 0)
	CPU_AddCommand($4E, "LSR", @LSR_M(), #Addr_Absolute,	6, 0, 0)
	CPU_AddCommand($4F, "SRE", @SRE(), #Addr_Absolute,		6, 0, 1); illegal
	CPU_AddCommand($50, "BVC", @BVC(), #Addr_Relative,		2, 0, 0)
	CPU_AddCommand($51, "EOR", @EOR(), #Addr_IIndirectY,	5, 1, 0)
	CPU_AddCommand($52, "KIL", @KIL(), #Addr_Implied,		1, 0, 1); illegal
	CPU_AddCommand($53, "SRE", @SRE(), #Addr_IndirectY,		8, 0, 1); illegal
	CPU_AddCommand($54, "NOP", @NOP(), #Addr_ZeropageX,		4, 0, 1); illegal
	CPU_AddCommand($55, "EOR", @EOR(), #Addr_IZeropageX,	4, 0, 0)	
	CPU_AddCommand($56, "LSR", @LSR_M(), #Addr_ZeropageX,	6, 0, 0)
	CPU_AddCommand($57, "SRE", @SRE(), #Addr_ZeropageX,		6, 0, 1); illegal
	CPU_AddCommand($58, "CLI", @CLI(), #Addr_Implied,		2, 0, 0)
	CPU_AddCommand($59, "EOR", @EOR(), #Addr_IAbsoluteY,	4, 1, 0)
	CPU_AddCommand($5A, "NOP", @NOP(), #Addr_Implied,		2, 0, 1); illegal
	CPU_AddCommand($5B, "SRE", @SRE(), #Addr_AbsoluteY,		7, 0, 1); illegal
	CPU_AddCommand($5C, "NOP", @NOP(), #Addr_AbsoluteX,		4, 0, 1); illegal
	CPU_AddCommand($5D, "EOR", @EOR(), #Addr_IAbsoluteX,	4, 1, 0)
	CPU_AddCommand($5E, "LSR", @LSR_M(), #Addr_AbsoluteX, 	7, 0, 0)
	CPU_AddCommand($5F, "SRE", @SRE(), #Addr_AbsoluteX,		7, 0, 1); illegal
	CPU_AddCommand($60, "RTS", @RTS(), #Addr_Implied,		6, 0, 0)
	CPU_AddCommand($61, "ADC", @ADC(), #Addr_IIndirectX,	6, 0, 0)
	CPU_AddCommand($62, "KIL", @KIL(), #Addr_Implied,		1, 0, 1); illegal
	CPU_AddCommand($63, "RRA", @RRA(), #Addr_IndirectX,		8, 0, 1); illegal
	CPU_AddCommand($64, "NOP", @NOP(), #Addr_Zeropage,		3, 0, 1); illegal
	CPU_AddCommand($65, "ADC", @ADC(), #Addr_IZeropage,		3, 0, 0)
	CPU_AddCommand($66, "ROR", @ROR_M(),#Addr_Zeropage,		5, 0, 0)
	CPU_AddCommand($67, "RRA", @RRA(), #Addr_Zeropage,		5, 0, 1); illegal
	CPU_AddCommand($68, "PLA", @PLA(), #Addr_Implied,		4, 0, 0)
	CPU_AddCommand($69, "ADC", @ADC(), #Addr_Immediate,		2, 0, 0)
	CPU_AddCommand($6A, "ROR", @ROR_A(), #Addr_Implied,		2, 0, 0)
	CPU_AddCommand($6B, "ARR", @ARR(), #Addr_Immediate,		2, 0, 0)
	CPU_AddCommand($6C, "JMP", @JMP(), #Addr_Indirect,		3, 0, 0)
	CPU_AddCommand($6D, "ADC", @ADC(), #Addr_IAbsolute,		4, 0, 0)
	CPU_AddCommand($6E, "ROR", @ROR_M(), #Addr_Absolute,	6, 0, 0)
	CPU_AddCommand($6F, "RRA", @RRA(), #Addr_Absolute,		6, 0, 1); illegal
	CPU_AddCommand($70, "BVS", @BVS(), #Addr_Relative,		2, 0, 0)
	CPU_AddCommand($71, "ADC", @ADC(), #Addr_IIndirectY,	5, 1, 0)
	CPU_AddCommand($72, "KIL", @KIL(), #Addr_Implied,		1, 0, 1); illegal
	CPU_AddCommand($73, "RRA", @RRA(), #Addr_IndirectY,		8, 0, 1); illegal
	CPU_AddCommand($74, "NOP", @NOP(), #Addr_ZeropageX,		4, 0, 1); illegal
	CPU_AddCommand($75, "ADC", @ADC(), #Addr_IZeropageX,	4, 0, 0)	
	CPU_AddCommand($76, "ROR", @ROR_M(), #Addr_ZeropageX,	6, 0, 0)
	CPU_AddCommand($77, "RRA", @RRA(), #Addr_ZeropageX,		6, 0, 1); illegal
	CPU_AddCommand($78, "SEI", @SEI(), #Addr_Implied,		2, 0, 0)
	CPU_AddCommand($79, "ADC", @ADC(), #Addr_IAbsoluteY,	4, 1, 0)
	CPU_AddCommand($7A, "NOP", @NOP(), #Addr_Implied,		2, 0, 1); illegal
	CPU_AddCommand($7B, "RRA", @RRA(), #Addr_IAbsoluteY,	7, 1, 1); illegal
	CPU_AddCommand($7C, "NOP", @NOP(), #Addr_AbsoluteX,		4, 0, 1); illegal
	CPU_AddCommand($7D, "ADC", @ADC(), #Addr_IAbsoluteX,	4, 1, 0)
	CPU_AddCommand($7E, "ROR", @ROR_M(), #Addr_AbsoluteX,	7, 0, 0)
	CPU_AddCommand($7F, "RRA", @RRA(), #Addr_AbsoluteX,		7, 0, 1); illegal
	CPU_AddCommand($80, "NOP", @NOP(), #Addr_Immediate,		2, 0, 0); illegal
	CPU_AddCommand($81, "STA", @STA(), #Addr_IndirectX,		6, 0, 0)
	CPU_AddCommand($82, "NOP", @NOP(), #Addr_Immediate,		2, 0, 1); illegal
	CPU_AddCommand($83, "SAX", @SAX(), #Addr_IndirectX,		6, 0, 1); illegal
	CPU_AddCommand($84, "STY", @STY(), #Addr_Zeropage,		3, 0, 0)
	CPU_AddCommand($85, "STA", @STA(), #Addr_Zeropage,		3, 0, 0)
	CPU_AddCommand($86, "STX", @STX(), #Addr_Zeropage,		3, 0, 0)
	CPU_AddCommand($87, "SAX", @SAX(), #Addr_Zeropage,		3, 0, 1); illegal
	CPU_AddCommand($88, "DEY", @DEY(), #Addr_Implied,		2, 0, 0)
	CPU_AddCommand($89, "NOP", @NOP(), #Addr_Immediate,		2, 0, 1); illegal
	CPU_AddCommand($8A, "TXA", @TXA(), #Addr_Implied,		2, 0, 0)
	CPU_AddCommand($8B, "XAA", @XAA(), #Addr_Immediate,		2, 0, 0)
	CPU_AddCommand($8C, "STY", @STY(), #Addr_Absolute,		4, 0, 0)
	CPU_AddCommand($8D, "STA", @STA(), #Addr_Absolute,		4, 0, 0)
	CPU_AddCommand($8E, "STX", @STX(), #Addr_Absolute,		4, 0, 0)
	CPU_AddCommand($8F, "SAX", @SAX(), #Addr_Absolute,		4, 0, 1); illegal
	CPU_AddCommand($90, "BCC", @BCC(), #Addr_Relative,		2, 0, 0)
	CPU_AddCommand($91, "STA", @STA(), #Addr_IndirectY,		5, 0, 0)
	CPU_AddCommand($92, "KIL", @KIL(), #Addr_Implied,		1, 0, 1); illegal
	CPU_AddCommand($93, "AXA", @AXA(), #Addr_IndirectY,		6, 0, 1); illegal
	CPU_AddCommand($94, "STY", @STY(), #Addr_ZeropageX,		4, 0, 0)
	CPU_AddCommand($95, "STA", @STA(), #Addr_ZeropageX,		4, 0, 0)
	CPU_AddCommand($96, "STX", @STX(), #Addr_ZeropageY,		4, 0, 0)
	CPU_AddCommand($97, "SAX", @SAX(), #Addr_ZeropageY,		4, 0, 1); illegal
	CPU_AddCommand($98, "TYA", @TYA(), #Addr_Implied,		2, 0, 0)
	CPU_AddCommand($99, "STA", @STA(), #Addr_AbsoluteY,		4, 0, 0)
	CPU_AddCommand($9A, "TXS", @TXS(), #Addr_Implied,		2, 0, 0)
	CPU_AddCommand($9B, "TAS", @TAS(), #Addr_AbsoluteY,		5, 0, 1); illegal
	CPU_AddCommand($9C, "SHY", @SHY(), #Addr_AbsoluteX,		5, 0, 1); illegal
	CPU_AddCommand($9D, "STA", @STA(), #Addr_AbsoluteX,		4, 0, 0)
	CPU_AddCommand($9E, "SHX", @SHX(), #Addr_AbsoluteY,		5, 0, 1); illegal
	CPU_AddCommand($9F, "AXA", @AXA(), #Addr_AbsoluteY,		5, 0, 1); illegal
	CPU_AddCommand($A0, "LDY", @LDY(), #Addr_Immediate,		2, 0, 0)
	CPU_AddCommand($A1, "LDA", @LDA(), #Addr_IIndirectX,	6, 0, 0)
	CPU_AddCommand($A2, "LDX", @LDX(), #Addr_Immediate,		2, 0, 0)
	CPU_AddCommand($A3, "LAX", @LAX(), #Addr_IndirectX,		6, 0, 1); illegal
	CPU_AddCommand($A4, "LDY", @LDY(), #Addr_IZeropage,		3, 0, 0)
	CPU_AddCommand($A5, "LDA", @LDA(), #Addr_IZeropage,		3, 0, 0)
	CPU_AddCommand($A6, "LDX", @LDX(), #Addr_IZeropage,		3, 0, 0)
	CPU_AddCommand($A7, "LAX", @LAX(), #Addr_Zeropage,		3, 0, 1); illegal
	CPU_AddCommand($A8, "TAY", @TAY(), #Addr_Implied,		2, 0, 0)
	CPU_AddCommand($A9, "LDA", @LDA(), #Addr_Immediate,		2, 0, 0)
	CPU_AddCommand($AA, "TAX", @TAX(), #Addr_Implied,		2, 0, 0)
	CPU_AddCommand($AB, "LAX", @LAX(), #Addr_Immediate,		2, 0, 1); illegal
	CPU_AddCommand($AC, "LDY", @LDY(), #Addr_IAbsolute,		4, 0, 0)
	CPU_AddCommand($AD, "LDA", @LDA(), #Addr_IAbsolute,		4, 0, 0)
	CPU_AddCommand($AE, "LDX", @LDX(), #Addr_IAbsolute,		4, 0, 0)
	CPU_AddCommand($AF, "LAX", @LAX(), #Addr_Absolute,		4, 0, 1); illegal
	CPU_AddCommand($B0, "BCS", @BCS(), #Addr_Relative,		2, 0, 0)
	CPU_AddCommand($B1, "LDA", @LDA(), #Addr_IIndirectY,	5, 1, 0)
	CPU_AddCommand($B2, "KIL", @KIL(), #Addr_Implied,		1, 0, 1); illegal
	CPU_AddCommand($B3, "LAX", @LAX(), #Addr_IndirectY,		5, 0, 1); illegal
	CPU_AddCommand($B4, "LDY", @LDY(), #Addr_IZeropageX,	4, 0, 0)	
	CPU_AddCommand($B5, "LDA", @LDA(), #Addr_IZeropageX,	4, 0, 0)	
	CPU_AddCommand($B6, "LDX", @LDX(), #Addr_IZeropageY,	4, 0, 0)	
	CPU_AddCommand($B7, "LAX", @LAX(), #Addr_ZeropageY,		4, 0, 1); illegal
	CPU_AddCommand($B8, "CLV", @CLV(), #Addr_Implied,		2, 0, 0)
	CPU_AddCommand($B9, "LDA", @LDA(), #Addr_IAbsoluteY,	4, 1, 0)
	CPU_AddCommand($BA, "TSX", @TSX(), #Addr_Implied,		2, 0, 0)
	CPU_AddCommand($BB, "LAS", @LAS(), #Addr_AbsoluteY,		4, 1, 1); illegal
	CPU_AddCommand($BC, "LDY", @LDY(), #Addr_IAbsoluteX,	4, 1, 0)
	CPU_AddCommand($BD, "LDA", @LDA(), #Addr_IAbsoluteX,	4, 1, 0)
	CPU_AddCommand($BE, "LDX", @LDX(), #Addr_IAbsoluteY,	4, 1, 0)
	CPU_AddCommand($BF, "LAX", @LAX(), #Addr_AbsoluteY,		4, 1, 1); illegal
	CPU_AddCommand($C0, "CPY", @CPY(), #Addr_Immediate,		2, 0, 0)
	CPU_AddCommand($C1, "CMP", @CMP(), #Addr_IIndirectX,	6, 0, 0)
	CPU_AddCommand($C2, "NOP", @NOP(), #Addr_Immediate,		2, 0, 1); illegal
	CPU_AddCommand($C3, "DCP", @DCP(), #Addr_IndirectX,		8, 0, 1); illegal
	CPU_AddCommand($C4, "CPY", @CPY(), #Addr_IZeropage,		3, 0, 0)
	CPU_AddCommand($C5, "CMP", @CMP(), #Addr_IZeropage,		3, 0, 0)
	CPU_AddCommand($C6, "DEC", @DEC(), #Addr_Zeropage,		5, 0, 0)
	CPU_AddCommand($C7, "DCP", @DCP(), #Addr_Zeropage,		5, 0, 1); illegal
	CPU_AddCommand($C8, "INY", @INY(), #Addr_Implied,		2, 0, 0)
	CPU_AddCommand($C9, "CMP", @CMP(), #Addr_Immediate,		2, 0, 0)
	CPU_AddCommand($CA, "DEX", @DEX(), #Addr_Implied,		2, 0, 0)
	CPU_AddCommand($CB, "SAX", @SAX(), #Addr_Immediate,		2, 0, 0)
	CPU_AddCommand($CC, "CPY", @CPY(), #Addr_IAbsolute,		4, 0, 0)
	CPU_AddCommand($CD, "CMP", @CMP(), #Addr_IAbsolute,		4, 0, 0)
	CPU_AddCommand($CE, "DEC", @DEC(), #Addr_Absolute,		6, 0, 0)
	CPU_AddCommand($CF, "DCP", @DCP(), #Addr_Absolute,		6, 0, 1); illegal
	CPU_AddCommand($D0, "BNE", @BNE(), #Addr_Relative,		2, 0, 0)
	CPU_AddCommand($D1, "CMP", @CMP(), #Addr_IIndirectY,	5, 1, 0)
	CPU_AddCommand($D2, "KIL", @KIL(), #Addr_Implied,		1, 0, 1); illegal
	CPU_AddCommand($D3, "DCP", @DCP(), #Addr_IndirectY,		8, 0, 1); illegal
	CPU_AddCommand($D4, "NOP", @NOP(), #Addr_ZeropageX,		4, 0, 1); illegal
	CPU_AddCommand($D5, "CMP", @CMP(), #Addr_IZeropageX,	4, 0, 0)
	CPU_AddCommand($D6, "DEC", @DEC(), #Addr_ZeropageX,		6, 0, 0)
	CPU_AddCommand($D7, "DCP", @DCP(), #Addr_ZeropageX,		6, 0, 1); illegal
	CPU_AddCommand($D8, "CLD", @CLD(), #Addr_Implied,		2, 0, 0)
	CPU_AddCommand($D9, "CMP", @CMP(), #Addr_IAbsoluteY,	4, 1, 0)
	CPU_AddCommand($DA, "NOP", @NOP(), #Addr_Implied,		2, 0, 1); illegal
	CPU_AddCommand($DB, "DCP", @DCP(), #Addr_AbsoluteY,		7, 0, 1); illegal
	CPU_AddCommand($DC, "NOP", @NOP(), #Addr_AbsoluteX,		4, 0, 1); illegal
	CPU_AddCommand($DD, "CMP", @CMP(), #Addr_IAbsoluteX,	4, 1, 0)
	CPU_AddCommand($DE, "DEC", @DEC(), #Addr_AbsoluteX,		7, 0, 0)
	CPU_AddCommand($DF, "DCP", @DCP(), #Addr_AbsoluteX,		7, 0, 1); illegal
	CPU_AddCommand($E0, "CPX", @CPX(), #Addr_Immediate,		2, 0, 0)
	CPU_AddCommand($E1, "SBC", @SBC(), #Addr_IIndirectX,	6, 0, 0)
	CPU_AddCommand($E2, "NOP", @NOP(), #Addr_Immediate,		2, 0, 1); illegal
	CPU_AddCommand($E3, "ISC", @ISC(), #Addr_IndirectX,		8, 0, 1); illegal
	CPU_AddCommand($E4, "CPX", @CPX(), #Addr_IZeropage,		3, 0, 0)
	CPU_AddCommand($E5, "SBC", @SBC(), #Addr_IZeropage,		3, 0, 0)
	CPU_AddCommand($E6, "INC", @INC(), #Addr_Zeropage,		5, 0, 0)
	CPU_AddCommand($E7, "ISC", @ISC(), #Addr_Zeropage,		5, 0, 1); illegal
	CPU_AddCommand($E8, "INX", @INX(), #Addr_Implied,		2, 0, 0)
	CPU_AddCommand($E9, "SBC", @SBC(), #Addr_Immediate,		2, 0, 0)
	CPU_AddCommand($EA, "NOP", @NOP(), #Addr_Implied,		2, 0, 0)
	CPU_AddCommand($EB, "SBC", @SBC(), #Addr_Immediate,		2, 0, 1); illegal
	CPU_AddCommand($EC, "CPX", @CPX(), #Addr_IAbsolute,		4, 0, 0)
	CPU_AddCommand($ED, "SBC", @SBC(), #Addr_IAbsolute,		4, 0, 0)
	CPU_AddCommand($EE, "INC", @INC(), #Addr_Absolute,		6, 0, 0)
	CPU_AddCommand($EF, "ISC", @ISC(), #Addr_Absolute,		6, 0, 1); illegal
	CPU_AddCommand($F0, "BEQ", @BEQ(), #Addr_Relative,		2, 0, 0)
	CPU_AddCommand($F1, "SBC", @SBC(), #Addr_IIndirectY,	5, 1, 0)
	CPU_AddCommand($F2, "KIL", @KIL(), #Addr_Implied,		1, 0, 1); illegal
	CPU_AddCommand($F3, "ISC", @ISC(), #Addr_IndirectY,		8, 0, 1); illegal
	CPU_AddCommand($F4, "NOP", @NOP(), #Addr_ZeropageX,		4, 0, 1); illegal
	CPU_AddCommand($F5, "SBC", @SBC(), #Addr_IZeropageX,	4, 0, 0)
	CPU_AddCommand($F6, "INC", @INC(), #Addr_ZeropageX,		6, 0, 0)
	CPU_AddCommand($F7, "ISC", @ISC(), #Addr_ZeropageX,		6, 0, 1); illegal
	CPU_AddCommand($F8, "SED", @SED(), #Addr_Implied,		2, 0, 0)
	CPU_AddCommand($F9, "SBC", @SBC(), #Addr_IAbsoluteY,	4, 1, 0)
	CPU_AddCommand($FA, "NOP", @NOP(), #Addr_Implied,		2, 0, 1); illegal
	CPU_AddCommand($FB, "ISC", @ISC(), #Addr_AbsoluteY,		7, 0, 1); illegal
	CPU_AddCommand($FC, "NOP", @NOP(), #Addr_AbsoluteX,		4, 0, 1); illegal
	CPU_AddCommand($FD, "SBC", @SBC(), #Addr_IAbsoluteX,	4, 1, 0)
	CPU_AddCommand($FE, "INC", @INC(), #Addr_AbsoluteX,		7, 0, 0)
	CPU_AddCommand($FF, "ISC", @ISC(), #Addr_AbsoluteX,		7, 0, 1); illegal
EndProcedure
; IDE Options = PureBasic 5.72 (Windows - x64)
; Folding = ---------------------
; EnableXP