﻿XIncludeFile("PB64__DECLARE.pbi")

Procedure MEMORY_LoadBin(path.s, *mem)
	Protected file = ReadFile(#PB_Any, path, #PB_File_SharedRead)
	Protected result
	
	If IsFile(file)
		result = ReadData(file, *mem, Lof(file))
		
		CloseFile(file)
	EndIf
	
	ProcedureReturn result
EndProcedure

Procedure MEMORY_Init()
	ClearStructure(@MEMORY, S_MEMORY)
	InitializeStructure(@MEMORY, S_MEMORY)
	
	MEMORY\BankLayout = $FF
	
	; load ROMs
	MEMORY_LoadBin("rom\basic.901226-01.bin", 		@MEMORY\BASICROM)
	MEMORY_LoadBin("rom\characters.901225-01.bin", 	@MEMORY\CHARROM)
	MEMORY_LoadBin("rom\kernal.901227-03.bin", 		@MEMORY\KernalROM)
	
	MEMORY_Poke(#AddressMemoryConfig,	$37)	
EndProcedure

Macro Memory_SetBank(page0_, page1_, page2_, page3_, page4_, page5_, page6_)
	MEMORY\Banks[0] = page0_
	MEMORY\Banks[1] = page1_
	MEMORY\Banks[2] = page2_
	MEMORY\Banks[3] = page3_
	MEMORY\Banks[4] = page4_
	MEMORY\Banks[5] = page5_
	MEMORY\Banks[6] = page6_
EndMacro

Procedure MEMORY_SetupBanks(v.a)
	If (MEMORY\BankLayout = v)
		ProcedureReturn
	EndIf

	Protected bankLayout.a = %11000 | (v & %111)
	Select bankLayout								;pg.0	pg.1	pg.2	pg.3	pg.4	pg.5	pg.6
		Case 31					:	Memory_SetBank(	#kRAM,	#kRAM,	#kRAM,	#kBas,	#kRAM,	#kIO,	#kKrn)
		Case 14,30				:	Memory_SetBank(	#kRAM,	#kRAM,	#kRAM,	#kRAM,	#kRAM,	#kIO,	#kKrn)
		Case 13,29,5			:	Memory_SetBank(	#kRAM,	#kRAM,	#kRAM,	#kRAM,	#kRAM,	#kIO,	#kRAM)
		Case 0,1,4,8,12,24,28	:	Memory_SetBank(	#kRAM,	#kRAM,	#kRAM,	#kRAM,	#kRAM,	#kRAM,	#kRAM)
		Case 27					:	Memory_SetBank(	#kRAM,	#kRAM,	#kRAM,	#kBas,	#kRAM,	#kChr,	#kKrn)
		Case 10,26				:	Memory_SetBank(	#kRAM,	#kRAM,	#kRAM,	#kRAM,	#kRAM,	#kChr,	#kKrn)
		Case 9,25				:	Memory_SetBank(	#kRAM,	#kRAM,	#kRAM,	#kRAM,	#kRAM,	#kChr,	#kRAM)
		Case 16 To 23			:	Memory_SetBank(	#kRAM,	#kNo,	#kcLo,	#kNo,	#kNo,	#kIO,	#kcHi)
		Case 15					:	Memory_SetBank(	#kRAM,	#kRAM,	#kcLo,	#kBas,	#kRAM,	#kIO,	#kKrn)
		Case 11					:	Memory_SetBank(	#kRAM,	#kRAM,	#kcLo,	#kBas,	#kRAM,	#kChr,	#kKrn)
		Case 7					:	Memory_SetBank(	#kRAM,	#kRAM,	#kcLo,	#kcHi,	#kRAM,	#kIO,	#kKrn)
		Case 6					:	Memory_SetBank(	#kRAM,	#kRAM,	#kRAM,	#kcHi,	#kRAM,	#kIO,	#kKrn)
		Case 3					:	Memory_SetBank(	#kRAM,	#kRAM,	#kcLo,	#kcHi,	#kRAM,	#kChr,	#kKrn)
		Case 2					:	Memory_SetBank(	#kRAM,	#kRAM,	#kRAM,	#kcHi,	#kRAM,	#kChr,	#kKrn)
	EndSelect
	
	MEMORY\BankLayout = v
EndProcedure

Procedure.u MEMORY_ReadW(ofs.u)
	ProcedureReturn MEMORY_Peek(ofs) | (MEMORY_Peek(ofs + 1) << 8)
EndProcedure

Procedure MEMORY_Poke(ofs.u, v.a)
	Select ofs
		Case #AddressMemoryConfig
 			MEMORY_SetupBanks(v)
	 		MEMORY\RAM[ofs] = v
 		Case $D000 To $DFFF
 			;page 5
 			If MEMORY\Banks[5] = #kIO
 				Select ofs
 					Case $D000 To $D3FF
 						VIC_WriteRegister(ofs & $3F, v)
 					Case $D400 To $D7FF
 						SID_WriteRegister(ofs & $1F, v)
 					Case $D800 To $DBFF
  						MEMORY\ColorRAM[ofs & $3FF] = v
 					Case $DC00 To $DCCC
 						CIA_WriteRegister(*CIA1, ofs & $F, v)
 					Case $DD00 To $DDFF
 						CIA_WriteRegister(*CIA2, ofs & $F, v)
 					Case $DE00 To $DEFF
;  						MEMORY\ExtROM[ofs - $D000] = v
 						; I/O Area #1, memory mapped registers or machine code routines of optional external devices (256 bytes). Layout and contents depend on the actual device.
 					Case $DF00 To $DFFF
; 						; I/O Area #2, memory mapped registers or machine code routines of optional external devices (256 bytes). Layout and contents depend on the actual device.
; 						MEMORY\ExtROM[ofs - $D000] = v
 				EndSelect
 			Else
 				MEMORY\RAM[ofs] = v
 			EndIf
 		Default
 			;page 0, 1, 2, 3, 4, 6
			MEMORY\RAM[ofs] = v
	EndSelect
EndProcedure

Procedure.a MEMORY_Peek(ofs.u)
	Protected v.a

	Select ofs
 		Case $0000 To $0FFF
 			;page 0
 			v = MEMORY\RAM[ofs]
 	 	Case $1000 To $7FFF
 	 		;page 1
 	 		If MEMORY\Banks[1] = #kRAM
 	 			v = MEMORY\RAM[ofs]
 	 		EndIf
 	 	Case $8000 To $9FFF
 	 		;page 2
 	 		If MEMORY\Banks[2] = #kRAM
 	 			v = MEMORY\RAM[ofs]
 	 		EndIf
 		Case $A000 To $BFFF
 			;page 3
 			If MEMORY\Banks[3] = #kRAM
 				v = MEMORY\RAM[ofs]
 			ElseIf MEMORY\Banks[3] = #kBas
 				v = MEMORY\BasicROM[ofs & $1FFF]
	 		EndIf
 		Case $C000 To $CFFF
 			;page 4
 			If MEMORY\Banks[4] = #kRAM
 				v = MEMORY\RAM[ofs]
	 		EndIf
 		Case $D000 To $DFFF
 			;page 5
 			If MEMORY\Banks[5] = #kIO
 				Select ofs
 					Case $D000 To $D3FF
  						v = VIC_ReadRegister(ofs & $3F)
 					Case $D400 To $D7FF
 						v = SID_ReadRegister(ofs & $1F)
 					Case $D800 To $DBFF
 						v = MEMORY\ColorRAM[ofs & $3FF]
 					Case $DC00 To $DCFF
 						v = CIA_ReadRegister(*CIA1, ofs & $F)
 					Case $DD00 To $DDFF
 						v = CIA_ReadRegister(*CIA2, ofs & $F)
 					Case $DE00 To $DEFF
;  						v = MEMORY\ExtROM[ofs & $3FFF]]
 						; I/O Area #1, memory mapped registers or machine code routines of optional external devices (256 bytes). Layout and contents depend on the actual device.
 					Case $DF00 To $DFFF
; 						; I/O Area #2, memory mapped registers or machine code routines of optional external devices (256 bytes). Layout and contents depend on the actual device.
; 						v = MEMORY\ExtROM[ofs & $3FFF]]
 				EndSelect
 			ElseIf MEMORY\Banks[5] = #kRAM
 				v = MEMORY\RAM[ofs]
 			ElseIf MEMORY\Banks[5] = #kChr 				
 				v = MEMORY\CharROM[ofs & $FFF]; - $D000]
 			EndIf
 		Case $E000 To $FFFF
 			;page 6
 			If MEMORY\Banks[6] = #kKrn
 				v = MEMORY\KernalROM[ofs & $1FFF]
 			ElseIf MEMORY\Banks[6] = #kRAM
 				v = MEMORY\RAM[ofs]
 			EndIf
	EndSelect
		
	ProcedureReturn v
EndProcedure
Procedure MEMORY_WriteByteNoIo(ofs.u, v.a)
	MEMORY\RAM[ofs] = v
EndProcedure

Procedure MEMORY_WriteWordNoIo(ofs.u, v.u)
 	MEMORY_WriteByteNoIo(ofs    , v & $FF)
 	MEMORY_WriteByteNoIo(ofs + 1, v >> 8)
EndProcedure

Procedure.a MEMORY_VIC_ReadB(ofs.u)
	Protected vic_addr.u = VIC\Mem_Base + (ofs & $3FFF)
	
	Select vic_addr
		Case $1000 To $1FFF:	ProcedureReturn MEMORY\CharROM[vic_addr - $1000]
		Case $9000 To $9FFF:	ProcedureReturn MEMORY\CharROM[vic_addr - $9000]
		Default:				ProcedureReturn MEMORY\RAM[vic_addr]
	EndSelect	
 EndProcedure
; IDE Options = PureBasic 5.72 (Windows - x64)
; CursorPosition = 148
; FirstLine = 133
; Folding = --
; EnableXP