﻿InitSprite()

XIncludeFile("PB64__DECLARE.pbi")
XIncludeFile("PB64_SDL.pbi")
XIncludeFile("PB64_CPU.pbi")
XIncludeFile("PB64_MEMORY.pbi")
XIncludeFile("PB64_VIC.pbi")
XIncludeFile("PB64_IO.pbi")
XIncludeFile("PB64_CIA.pbi")
XIncludeFile("PB64_SID.pbi")

Procedure C64_Reset()
	
	MEMORY_Init()
	CPU_Init()
	CIA_Init(@CIA1)
	CIA_Init(@CIA2)
	VIC_Init()
	SID_Init()
	IO_Init()
	
	CPU_Reset()
	
	C64\State = 0
	
	StatusBarText(#StatusBar, 0, "*ready*")
EndProcedure

Procedure C64_Emulate()
	Protected cycles, n
	
	cycles = CPU_Emulate()
	For n = 1 To cycles
		If SID_Emulate() = #False
			ProcedureReturn #False
		EndIf
		If VIC_Emulate() = #False
			ProcedureReturn #False
		EndIf
	Next
	
	If CIA_Emulate(*CIA1, cycles) = #False
		ProcedureReturn #False
	EndIf
	If CIA_Emulate(*CIA2, cycles) = #False
		ProcedureReturn #False
	EndIf		
	
	ProcedureReturn #True
EndProcedure

Procedure C64_Start()
	C64_Reset()
	
	Repeat
  		C64_Emulate()
		
  		If C64\State & #C64_State_Error
  			If MessageRequester("Error",  C64\ErrorMessage + #CRLF$ + "Restart?", #PB_MessageRequester_Error | #PB_MessageRequester_YesNo) = #PB_MessageRequester_Yes
  				C64_Reset()
  			Else
  				C64\State | #C64_State_Quit
  			EndIf
		EndIf
	Until C64\State & #C64_State_Quit
	
	SID_Quit()
	End
EndProcedure

Procedure C64_Test(testFile.s, ofs.u, startAddress.u)
	C64_Reset()
	
 	MEMORY_Poke(#AddressMemoryConfig, 0)
	
	If MEMORY_LoadBin(testFile, @MEMORY\RAM[ofs]) = 0
		MessageRequester("", "file not found: " + GetFilePart(testFile))
	Else
		CPU\PC = startAddress
;  	 	CPU_Debug(CPU\PC, $FFFF, "d:\temp\test.txt", 0):End
		
		Protected pc = CPU\PC
		Repeat
			If CPU_Emulate() = #False
				MessageRequester("", "CPU error")
				Break
			EndIf
			If CPU\PC = $3463
				MessageRequester("", "test passed!")
				Break
			ElseIf pc = CPU\PC
				MessageRequester("", "infinite loop at " + Hex(pc))
				Break
			EndIf
			pc = CPU\PC
		ForEver
	EndIf
EndProcedure

Procedure C64_MainWindow()
	Protected i
	
  	If OpenWindow(#MainWindow, 0, 0, #VisibleScreenWidth * C64\Zoom, #VisibleScreenHeight * C64\Zoom + MenuHeight(), "PB64", #PB_Window_SystemMenu | #PB_Window_SizeGadget | #PB_Window_MaximizeGadget | #PB_Window_MinimizeGadget) = 0
		MessageRequester("", "OpenWindow failed", #PB_MessageRequester_Error)
		End
	EndIf
	If CreateStatusBar(#StatusBar, WindowID(#MainWindow)) = 0
		MessageRequester("", "CreateStatusBar failed", #PB_MessageRequester_Error)
		End
	Else
		AddStatusBarField(WindowWidth(#MainWindow) - 100)
		AddStatusBarField(100)
	EndIf
	ResizeWindow(#MainWindow, #PB_Ignore, #PB_Ignore, #PB_Ignore, WindowHeight(#MainWindow) + StatusBarHeight(#StatusBar))
	
  	If OpenWindowedScreen(WindowID(#MainWindow), 0, 0, #VisibleScreenWidth, #VisibleScreenHeight, 1, 0, StatusBarHeight(#StatusBar), #PB_Screen_NoSynchronization) = 0
		MessageRequester("", "OpenWindowedScreen failed", #PB_MessageRequester_Error)
	 	End	
	 EndIf
	
	If CreateMenu(#Menu_0, WindowID(#MainWindow)) = 0
		MessageRequester("", "CreateMenu failed", #PB_MessageRequester_Error)
		End
	Else
		MenuTitle("File")
		MenuItem(#m_Open, "Load...")		
		OpenSubMenu("Recent Files")
		For i = #m_QuickLoadFirst To #m_QuickLoadLast
			MenuItem(#m_QuickLoadFirst + i, "-")
		Next
		CloseSubMenu()
		MenuItem(#m_AutoRun, "Autostart")

		MenuBar()
		MenuItem(#m_SaveState, "Save State")
		MenuItem(#m_LoadState, "Load State")
		MenuBar()
		MenuItem(#m_PasteClip, "Paste Clipboard")
		MenuBar()
		MenuItem(#m_Quit, "Quit")
		MenuBar()
		MenuTitle("Emulation")
		MenuItem(#m_Joystickport1, "Joystickport: 1")
		MenuItem(#m_Joystickport2, "Joystickport: 2")
		OpenSubMenu("Settings")
			MenuItem(#m_JoystickKeys, "Enabled")
			OpenSubMenu("Set Keys")
				MenuItem(#m_JoystickConfigUp,"Up")
				MenuItem(#m_JoystickConfigDown, "Down")
				MenuItem(#m_JoystickConfigLeft, "Left")
				MenuItem(#m_JoystickConfigRight, "Right")
				MenuItem(#m_JoystickConfigFire, "Fire")
			CloseSubMenu()
		CloseSubMenu()
		MenuBar()
		MenuItem(#m_Vsync, "Vsync")
		MenuBar()
		MenuItem(#m_Sound, "Sound")
		MenuBar()
		MenuItem(#m_Reset, "Reset")
		
		SetMenuItemState(#Menu_0, #m_AutoRun, 1)
		If c64\JoystickPort = #C64_Joystick_Port1
			SetMenuItemState(#Menu_0, #m_Joystickport1, 1)
		Else
			SetMenuItemState(#Menu_0, #m_Joystickport2, 1)
		EndIf		
		SetMenuItemState(#Menu_0, #m_Vsync, C64\vSyncEnabled)
		SetMenuItemState(#Menu_0, #m_Sound, C64\SoundEnabled)
		SetMenuItemState(#Menu_0, #m_JoystickKeys, 0)
		
		If OpenPreferences(PreferenceFile)
			PreferenceGroup("Quickload")
			For i = #m_QuickLoadFirst To #m_QuickLoadLast
				SetMenuItemText(#Menu_0, i, GetFilePart(ReadPreferenceString("Recent" + Str(i), "-")))
			Next
			
			PreferenceGroup("JoystickKeys")
			C64\JoystickKey[#C64_Joystick_Up]		= ReadPreferenceInteger("JoystickUp", 		#PB_Key_Up)
			C64\JoystickKey[#C64_Joystick_Down]		= ReadPreferenceInteger("JoystickDown", 	#PB_Key_Down)
			C64\JoystickKey[#C64_Joystick_Left]		= ReadPreferenceInteger("JoystickLeft", 	#PB_Key_Left)
			C64\JoystickKey[#C64_Joystick_Right]	= ReadPreferenceInteger("JoystickRight", 	#PB_Key_Right)
			C64\JoystickKey[#C64_Joystick_Fire]		= ReadPreferenceInteger("JoystickFire", 	#PB_Key_Space)
			
			ClosePreferences()		
		EndIf		
	EndIf
			 
	RemoveKeyboardShortcut(#MainWindow, #PB_Shortcut_All)
	KeyboardMode(#PB_Keyboard_International)
EndProcedure

DisableExplicit

C64_MainWindow()
;C64_Test("test\6502_functional_test2.bin", $400, $400): End
C64_Start()

; IDE Options = PureBasic 5.72 (Windows - x64)
; CursorPosition = 195
; FirstLine = 144
; Folding = -
; EnableXP